import xmlrpc, { createClient, createSecureClient } from 'xmlrpc';
import url from 'url';

export type TOdooConfig = {
  url: string;
  port: string | number;
  db: string;
  username: string;
  password: string;
};

export class Odoo {
  // eslint-disable-next-line @typescript-eslint/ban-ts-comment
  // @ts-ignore
  private config: TOdooConfig;
  private host: string;
  private port: string | number;
  private db: string;
  private username: string;
  private password: string;
  private secure: boolean;
  private uid: number;

  constructor(config: TOdooConfig) {
    const urlParts = url.parse(config.url);
    this.config = config;
    this.host = urlParts.hostname!!;
    this.port = config.port || urlParts.port!!;
    this.db = config.db;
    this.username = config.username;
    this.password = config.password;
    this.secure = true;
    if (urlParts.protocol !== 'https:') {
      this.secure = false;
    }
    this.uid = 0;
  }

  private _getClient(path: string): xmlrpc.Client {
    if (this.secure) {
      return createSecureClient({
        host: this.host,
        port: Number(this.port),
        path,
      });
    }
    return createClient({
      host: this.host,
      port: Number(this.port),
      path,
    });
  }

  private _methodCall<T = any>(
    client: xmlrpc.Client,
    method: string,
    params: any[] = []
  ): Promise<T> {
    return new Promise((resolve, reject) => {
      client.methodCall(method, params, (err, value) => {
        if (err) {
          console.log(err);
          return reject(err);
        }
        return resolve(value);
      });
    });
  }

  public connect<T = any>(): Promise<T> {
    var client = this._getClient('/xmlrpc/2/common');
    var params: any[] = [];
    params.push(this.db);
    params.push(this.username);
    params.push(this.password);
    params.push({});
    return new Promise((resolve, reject) => {
      client.methodCall('authenticate', params, (error, value) => {
        if (error) {
          reject(error);
        }
        if (!value) {
          reject(new Error('No UID returned from authentication.'));
        }
        this.uid = value;
        resolve(value);
      });
    });
  }

  public async execute_kw<T = any>(
    model: string,
    method: string,
    params: any[] = []
  ): Promise<T> {
    const client = this._getClient('/xmlrpc/2/object');
    const _params: any[] = [];
    _params.push(this.db);
    _params.push(this.uid);
    _params.push(this.password);
    _params.push(model);
    _params.push(method);
    _params.push(...params);
    const value = await this._methodCall<T>(client, 'execute_kw', _params);
    return value;
  }

  public async exec_workflow<T = any>(
    model: string,
    method: string,
    params: any[] = []
  ) {
    const client = this._getClient('/xmlrpc/2/object');
    const _params: any[] = [];
    _params.push(this.db);
    _params.push(this.uid);
    _params.push(this.password);
    _params.push(model);
    _params.push(method);
    _params.push(...params);
    const value = await this._methodCall<T>(client, 'exec_workflow', _params);
    return value;
  }

  public async render_report<T = any>(
    report: string,
    params: any[] = []
  ): Promise<T> {
    const client = this._getClient('/xmlrpc/2/report');
    const _params = [];
    _params.push(this.db);
    _params.push(this.uid);
    _params.push(this.password);
    _params.push(report);
    _params.push(...params);
    const value = await this._methodCall<T>(client, 'render_report', _params);
    return value;
  }
}
